; SketchyLISP Library
; Copyright (C) 2005 Nils M Holm. All rights reserved.
; See the file LICENSE of the SketchyLISP distribution
; for conditions of use.

; ---name---
; n+

; ---conformance---
; Sketchy Core

; ---purpose---
; Add two natural numbers.

; ---args---
; A - natural number
; B - natural number

; ---keywords---
; N+ function, plus, sum, addition, arithmetics, natural

; ---see-also---
; digits, n-, nquotient, nremainder, n*, +

; ---example---
; (n+ 5 7) => 12

(define nplus #t)

:c lib/digits.l 0d 1d
:c lib/reverse.l

; ---code---
(define n+ (lambda (a b)
	(letrec
	
		; Add two lists of decimal digits forming reverse
		; decimal numbers. The result is in normal order.
		; A B - reverse lists of digits
		; C - carry flag
		; R - result (normal order)
		((add (lambda (a b c r)
			(cond ((null? a)
					(cond
						; A=() ; B=()
						((null? b)
							(cond
								; carry not set? return R
								((eq? c 0d) r)	
								; else prepend overflow
								(#t (cons 1d r))))
						; B\=(): keep adding leading zeroes
						(#t (add '() (cdr b)
							(cdr (d+ 0d (car b) c))
							(cons (car (d+ 0d (car b) c))
								r)))))
			((null? b)
				; B=() ; A\=()
				(add (cdr a) '()
					(cdr (d+ (car a) 0d c))
					(cons (car (d+ (car a) 0d c)) r)))
			; default: advance to next two digits
			(#t (add (cdr a) (cdr b)
				(cdr (d+ (car a) (car b) c))
				(cons (car (d+ (car a) (car b) c)) r)))))))

		(list->integer
			(add (reverse (integer->list a))
				(reverse (integer->list b))
				0d '()) #t))))

