
(define eval (lambda (x e)
	(letrec

	; Pre-defined environment
	((initial-env
		'((#f.#f)
		(#t.#t)
		(bottom.bottom)
		(car.car)
		(cdr.cdr)
		(cond.cond)
		(cons.cons)
		(eq?.eq?)
		(lambda.lambda)
		(%lambda.%lambda)
		(letrec.letrec)
		(null?.null?)
		(pair?.pair?)
		(procedure?.procedure?)
		(quote.quote)))

	; Evaluate each member of list X in environment E:
	; (M1 ... Mn) -> (eval[M1] ... eval[Mn])
	(evlist (lambda (x e)
		(cond ((null? x) '())
			(#t (cons (_eval (car x) e) (evlist (cdr x) e))))))

	; Find the value of a symbol
	(value (lambda (x e)
		(letrec ((b (assoc x e)))
			(cond ((null? b) (bottom eval:symbol-not-bound: x))
				(#t (cdr b))))))

	; Create a LETREC* environment
	(makeenv (lambda (bindings env)
		(letrec

			; Evaluate and bind arguments.
			((make-bindings (lambda (bindings new-env)
				(cond ((null? bindings) new-env)
					(#t (make-bindings (cdr bindings)
						(cons (cons
								(caar bindings)
								; Evaluate arguments in the union of
								; the outer environment and the local
								; environment built so far.
								(_eval (cadar bindings)
									(append new-env env)))
							new-env)))))))

			(append (make-bindings bindings '()) env))))

	; Create closure from LAMBDA function
	(make-fun (lambda (args body env)
			(list '%lambda args body)))

	; Select first clause of COND that has a true predicate.
	(do-cond (lambda (clauses env)
		(cond ((null? clauses)
				(bottom 'eval:cond-no-default))
			(#t (letrec ((tval (_eval (caar clauses) env)))
				(cond ((eq? tval #t)
						(cadar clauses))
					((eq? tval #f)
						(do-cond (cdr clauses) env))
					(#t (bottom 'eval:cond-no-tval tval))))))))

	; Is X a procedure?
	(procedure-p (lambda (x)
		(cond ((pair? x) (eq? (car x) '%lambda))
			(#t #f))))

	; Evaluate X in E
	(_eval (lambda (x e)
		(cond ((symbol? x)
				(value x e))
			((procedure-p x)
				x)
			((boolean? x)
				x)
			((null? x)
				'())
			(#t (letrec ((fun (_eval (car x) e)))
				(cond ((eq? fun 'bottom)
						(list (append 'bottom: (cdr x))))
					((eq? fun 'car)
						(car (_eval (cadr x) e)))
					((eq? fun 'cdr)
						(cdr (_eval (cadr x) e)))
					((eq? fun 'cond)
						(_eval (do-cond (cdr x) e) e))
					((eq? fun 'cons)
						(cons (_eval (cadr x) e) (_eval (caddr x) e)))
					((eq? fun 'eq?)
						(eq? (_eval (cadr x) e) (_eval (caddr x) e)))
					((eq? fun 'lambda)
						(make-fun (cadr x) (caddr x) e))
					((eq? fun 'letrec)
						(_eval (caddr x) (makeenv (cadr x) e)))
					((eq? fun 'null?)
						(null? (_eval (cadr x) e)))
					((eq? fun 'pair?)
						(pair? (_eval (cadr x) e)))
					((eq? fun 'procedure?)
						(procedure-p (_eval (cadr x) e)))
					((eq? fun 'quote)
						(cadr x))
					((symbol? fun)
						(bottom (list 'eval:non-function fun)))
					((null? fun)
						(bottom (list 'eval:non-function (car x))))
					((eq? (car fun) '%lambda)
						(_eval (caddr fun)
							(append
								(pair (cadr fun) (evlist (cdr x) e)) e)))
					(#t (bottom (list 'eval:non-function fun))))))))))

	(_eval x (append e initial-env)))))

